#
#  Makefile Generator
#
package Generator;
use strict;
use Util qw(util_split diag_error util_format_variable);

# Generator->new($pParsedFile)
#    Create new generator
sub new {
    my $proto = shift;
    my $file = shift;
    my $expander = Expander->new($file);
    $expander->setExpandExpandable(1);

    bless { file => $file,
            outVars => "",
            outRules => "",
            outPat => "",
            didVars => {},
            didRules => {},
            expander => $expander,
            minLength => 100,
            wrapLength => 75,
            include => 'include' },
      ref($proto) || $proto;
}

# Include command
sub setInclude {
    my $self = shift;
    my $cmd = shift;
    $self->{include} = $cmd;
}

# $pGenerator->outputVariable(name)
#     Make sure that variable 'name' is mentioned in output.
sub outputVariable {
    my $self = shift;
    my $name = shift;

    # Skip if we already did it
    if (exists($self->{didVars}{$name})) { return }
    $self->{didVars}{$name} = 1;

    # Skip if it does not exist
    my $var = $self->{file}->getVar($name);
    if (!defined($var)) { return }

    # Skip if it is an internal variable
    if ($var->isExpand()) { return }

    # Okay, we want to output it. First print all referenced variables.
    my $value = $self->{expander}->expandVars($var->getStringValue());
    $self->processVars($value);

    # Now the variable
    $self->{outVars} .= util_format_variable($name, '=', $value,
                                             $self->{minLength},
                                             $self->{wrapLength});
}

# Add a blank line after current list of variables
sub addSpaceAfterVars {
    shift->{outVars} .= "\n";
}

# $pGenerator->outputRules(name, name, ....)
#      Generate rules
sub outputRules {
    my $self = shift;
    my @todo = @_;

    while (@todo) {
        my $name = shift @todo;

        # Skip if we already did it
        if (exists($self->{didRules}{$name})) { next }
        $self->{didRules}{$name} = 1;

        # Skip if it does not exist
        # xref ParsedFile
        if (!exists $self->{file}{rule}{$name}) { next }
        my $rule = $self->{file}{rule}{$name};

        # Output it
        my $expandedName = $self->{expander}->expandVars($name);

        my $accum = "$expandedName:";
        my @todo;
        foreach (@{$rule->getDepend()}) {
            push @todo, $_;
            my $now = $self->{expander}->expandVars($_);
            $self->processVars($now);
            foreach (util_split($now)) {
                # FIXME: handle user setting for wrap
                if (length($accum) + length($_) > 75) {
                    $self->{outRules} .= "$accum \\\n";
                    $accum = "   ";
                }
                $accum .= " $_";
            }
        }
        $self->{outRules} .= "$accum\n";
        foreach (@{$rule->getCommands()}) {
            my $now = $self->{expander}->expandVars($_);
            $self->processVars($now);
            $self->{outRules} .= "\t$now\n";
        }
        $self->{outRules} .= "\n";
    }
}

sub outputPatternRules {
    # xref ParsedFile, Rule
    my $self = shift;
    my $pPatHash = $self->{file}{pat};
    foreach my $input (sort keys %$pPatHash) {
        foreach my $output (sort keys %{$pPatHash->{$input}}) {
            if ($pPatHash->{$input}{$output}{used}) {
                $self->{outPat} .= ".$input.$output:\n";
                foreach (@{$pPatHash->{$input}{$output}->getCommands()}) {
                    my $now = $self->{expander}->expandVars($_);
                    $self->processVars($now);
                    $self->{outPat} .= "\t$now\n";
                }
                $self->{outPat} .= "\n";
            }
        }
    }
}

sub processVars {
    my $self = shift;
    my $value = shift;
    while ($value =~ m/\$\((.*?)\)/g) {
        $self->outputVariable($1);
    }
    while ($value =~ m/\$([^$(])/g) {
        $self->outputVariable($1);
    }
}

sub outputToFile {
    my $self = shift;
    my $name = shift;
    open OUT, "> $name" or
      diag_perror($name);
    print OUT
      "#\n# Makefile\n#\n\n",
        $self->{outVars},
          "\n";
    if (exists $self->{file}{includevar}) {
        foreach (@{$self->{file}{includevar}}) {
            print OUT "$self->{include} $_\n";
        }
        print OUT "\n";
    }

    print OUT $self->{outRules}, "\n";
    if (exists $self->{file}{includerule}) {
        foreach (@{$self->{file}{includerule}}) {
            print OUT "$self->{include} $_\n";
        }
        print OUT "\n";
    }

    print OUT $self->{outPat}, "\n";

    close OUT;
}

1;
